# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Provides custom Pythonic errors for dm_env_rpc error messages."""
from google.rpc import status_pb2


class DmEnvRpcError(Exception):
  """A dm_env_rpc custom exception.

  Wraps a google.rpc.Status message as a Python Exception class.
  """

  def __init__(self, status_proto: status_pb2.Status):
    super().__init__()
    self._status_proto = status_proto

  @property
  def code(self) -> int:
    return self._status_proto.code

  @property
  def message(self) -> str:
    return self._status_proto.message

  def __str__(self):
    return str(self._status_proto)

  def __reduce__(self):
    return (DmEnvRpcError, (self._status_proto,))



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""A base class for CreateWorld and DestroyWorld tests for a server."""
import abc

from absl.testing import absltest

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import error


class CreateDestroyWorld(absltest.TestCase, metaclass=abc.ABCMeta):
  """A base class for `CreateWorld` and `DestroyWorld` compliance tests."""

  @abc.abstractproperty
  def required_world_settings(self):
    """A string to Tensor mapping of the minimum set of required settings."""
    pass

  @abc.abstractproperty
  def invalid_world_settings(self):
    """World creation settings which are invalid in some way."""
    pass

  @abc.abstractproperty
  def has_multiple_world_support(self):
    """Does the server support creating more than one world?"""
    pass

  @abc.abstractproperty
  def connection(self):
    """An instance of dm_env_rpc's Connection."""
    pass

  def create_world(self, settings):
    """Returns the world name of the world created with the given settings."""
    response = self.connection.send(
        dm_env_rpc_pb2.CreateWorldRequest(settings=settings))
    return response.world_name

  def destroy_world(self, world_name):
    """Destroys the world named `world_name`."""
    if world_name is not None:
      self.connection.send(
          dm_env_rpc_pb2.DestroyWorldRequest(world_name=world_name))

  # pylint: disable=missing-docstring
  def test_can_create_and_destroy_world(self):
    # If this doesn't raise an exception the test passes.
    world_name = self.create_world(self.required_world_settings)
    self.destroy_world(world_name)

  def test_cannot_create_world_with_less_than_required_settings(self):
    settings = self.required_world_settings

    for name, _ in settings.items():
      sans_setting = dict(settings)
      del sans_setting[name]
      with self.assertRaises(error.DmEnvRpcError):
        self.create_world(sans_setting)

  def test_cannot_create_world_with_invalid_settings(self):
    settings = self.required_world_settings
    invalid_settings = self.invalid_world_settings
    for name, tensor in invalid_settings.items():
      with self.assertRaises(error.DmEnvRpcError):
        self.create_world({name: tensor, **settings})

  def test_world_name_is_unique(self):
    if not self.has_multiple_world_support:
      return
    world1_name = None
    world2_name = None
    try:
      world1_name = self.create_world(self.required_world_settings)
      world2_name = self.create_world(self.required_world_settings)
      self.assertIsNotNone(world1_name)
      self.assertIsNotNone(world2_name)
      self.assertNotEqual(world1_name, world2_name)
    finally:
      self.destroy_world(world1_name)
      self.destroy_world(world2_name)

  def test_cannot_destroy_uncreated_world(self):
    with self.assertRaises(error.DmEnvRpcError):
      self.destroy_world('foo')
  # pylint: enable=missing-docstring



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for dm_env_rpc error module."""

import pickle

from absl.testing import absltest
from google.rpc import code_pb2
from google.rpc import status_pb2
from dm_env_rpc.v1 import error


class ErrorTest(absltest.TestCase):

  def testSimpleError(self):
    message = status_pb2.Status(
        code=code_pb2.INVALID_ARGUMENT, message='A test error.')
    exception = error.DmEnvRpcError(message)

    self.assertEqual(code_pb2.INVALID_ARGUMENT, exception.code)
    self.assertEqual('A test error.', exception.message)
    self.assertEqual(str(message), str(exception))

  def testPickleUnpickle(self):
    exception = error.DmEnvRpcError(status_pb2.Status(
        code=code_pb2.INVALID_ARGUMENT, message='foo.'))
    pickled = pickle.dumps(exception)
    unpickled = pickle.loads(pickled)

    self.assertEqual(code_pb2.INVALID_ARGUMENT, unpickled.code)
    self.assertEqual('foo.', unpickled.message)


if __name__ == '__main__':
  absltest.main()



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for dm_env_flatten_utils."""

import collections

from absl.testing import absltest
from dm_env_rpc.v1 import dm_env_flatten_utils


class FlattenUtilsTest(absltest.TestCase):

  def test_flatten(self):
    input_dict = {
        'foo': {
            'bar': 1,
            'baz': False
        },
        'fiz': object(),
    }
    expected = {
        'foo.bar': 1,
        'foo.baz': False,
        'fiz': object(),
    }
    self.assertSameElements(expected,
                            dm_env_flatten_utils.flatten_dict(input_dict, '.'))

  def test_unflatten(self):
    input_dict = {
        'foo.bar.baz': True,
        'fiz.buz': 1,
        'foo.baz': 'val',
        'buz': {}
    }
    expected = {
        'foo': {
            'bar': {
                'baz': True
            },
            'baz': 'val'
        },
        'fiz': {
            'buz': 1
        },
        'buz': {},
    }
    self.assertSameElements(
        expected, dm_env_flatten_utils.unflatten_dict(input_dict, '.'))

  def test_unflatten_different_separator(self):
    input_dict = {'foo::bar.baz': True, 'foo.bar::baz': 1}
    expected = {'foo': {'bar.baz': True}, 'foo.bar': {'baz': 1}}
    self.assertSameElements(
        expected, dm_env_flatten_utils.unflatten_dict(input_dict, '::'))

  def test_flatten_unflatten(self):
    input_output = {
        'foo': {
            'bar': 1,
            'baz': False
        },
        'fiz': object(),
    }
    self.assertSameElements(
        input_output,
        dm_env_flatten_utils.unflatten_dict(
            dm_env_flatten_utils.flatten_dict(input_output, '.'), '.'))

  def test_flatten_with_key_containing_separator_raises_error(self):
    with self.assertRaisesRegex(ValueError, 'foo.bar'):
      dm_env_flatten_utils.flatten_dict({'foo.bar': True}, '.')

  def test_invalid_flattened_dict_raises_error(self):
    input_dict = collections.OrderedDict((
        ('foo.bar', True),
        ('foo', 'invalid_value_for_sub_key'),
    ))
    with self.assertRaisesRegex(ValueError, 'Duplicate key'):
      dm_env_flatten_utils.unflatten_dict(input_dict, '.')

  def test_sub_tree_has_value_raises_error(self):
    input_dict = collections.OrderedDict((
        ('branch', 'should_not_have_value'),
        ('branch.leaf', True),
        ))
    with self.assertRaisesRegex(ValueError,
                                "Sub-tree 'branch' has already been assigned"):
      dm_env_flatten_utils.unflatten_dict(input_dict, '.')

  def test_empty_dict_values_flatten(self):
    input_dict = {
        'foo': {},
        'bar': {
            'baz': {}
        },
    }
    expected = {
        'foo': {},
        'bar.baz': {},
    }
    self.assertSameElements(expected,
                            dm_env_flatten_utils.flatten_dict(input_dict, '.'))


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""An implementation of a dm_env environment using dm_env_rpc."""

import collections
from typing import Any, Iterable, Mapping, Optional, Sequence, Tuple
import dm_env
import immutabledict

from dm_env_rpc.v1 import connection as dm_env_rpc_connection
from dm_env_rpc.v1 import dm_env_flatten_utils
from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import dm_env_utils
from dm_env_rpc.v1 import error
from dm_env_rpc.v1 import spec_manager
from dm_env_rpc.v1 import tensor_utils

# Default observation names for common RL concepts.  By default the dm_env
# wrapper will use these for reward and discount if available, but this behavior
# can be overridden.
DEFAULT_REWARD_KEY = 'reward'
DEFAULT_DISCOUNT_KEY = 'discount'

# Default key separator, used in flattening/unflattening nested structures.
DEFAULT_KEY_SEPARATOR = '.'


class DmEnvAdaptor(dm_env.Environment):
  """An implementation of dm_env using dm_env_rpc as the data protocol.

  Users can also optionally provide a mapping of objects to DmEnvAdaptor
  attributes. This is to accomodate user-created protocol extensions that
  compliment the core protocol.

  """

  # Disable pytype attribute checking for dynamically created extension attrs.
  _HAS_DYNAMIC_ATTRIBUTES = True

  def __init__(
      self,
      connection: dm_env_rpc_connection.Connection,
      specs: dm_env_rpc_pb2.ActionObservationSpecs,
      requested_observations: Optional[Sequence[str]] = None,
      nested_tensors: bool = True,
      extensions: Optional[Mapping[str, Any]] = immutabledict.immutabledict()):
    """Initializes the environment with the provided dm_env_rpc connection.

    Args:
      connection: An instance of Connection already connected to a dm_env_rpc
        server and after a successful JoinWorldRequest has been sent.
      specs: A dm_env_rpc ActionObservationSpecs message for the environment.
      requested_observations: List of observation names to be requested from the
        environment when step is called. If None is specified then all
        observations will be requested.
      nested_tensors: Boolean to determine whether to flatten/unflatten tensors.
      extensions: Optional mapping of extension instances to DmEnvAdaptor
        attributes. Raises ValueError if attribute already exists.
    """
    self._dm_env_rpc_specs = specs
    self._action_specs = spec_manager.SpecManager(specs.actions)
    self._observation_specs = spec_manager.SpecManager(specs.observations)
    self._connection = connection
    self._last_state = dm_env_rpc_pb2.EnvironmentStateType.TERMINATED
    self._nested_tensors = nested_tensors

    if requested_observations is None:
      requested_observations = self._observation_specs.names()
      self._is_reward_requested = False
      self._is_discount_requested = False
    else:
      self._is_reward_requested = DEFAULT_REWARD_KEY in requested_observations
      self._is_discount_requested = (
          DEFAULT_DISCOUNT_KEY in requested_observations)

    requested_observations = set(requested_observations)

    self._default_reward_spec = None
    self._default_discount_spec = None
    if DEFAULT_REWARD_KEY in self._observation_specs.names():
      self._default_reward_spec = dm_env_utils.tensor_spec_to_dm_env_spec(
          self._observation_specs.name_to_spec(DEFAULT_REWARD_KEY))
      requested_observations.add(DEFAULT_REWARD_KEY)
    if DEFAULT_DISCOUNT_KEY in self._observation_specs.names():
      self._default_discount_spec = (
          dm_env_utils.tensor_spec_to_dm_env_spec(
              self._observation_specs.name_to_spec(DEFAULT_DISCOUNT_KEY)))
      requested_observations.add(DEFAULT_DISCOUNT_KEY)

    unsupported_observations = requested_observations.difference(
        self._observation_specs.names())
    if unsupported_observations:
      raise ValueError('Unsupported observations requested: {}'.format(
          unsupported_observations))
    self._requested_observation_uids = [
        self._observation_specs.name_to_uid(name)
        for name in requested_observations
    ]

    # Not strictly necessary but it makes the unit tests deterministic.
    self._requested_observation_uids.sort()

    if extensions is not None:
      for extension_name, extension in extensions.items():
        if hasattr(self, extension_name):
          raise ValueError(
              f'DmEnvAdaptor already has attribute "{extension_name}"!')
        setattr(self, extension_name, extension)

  def reset(self):
    """Implements dm_env.Environment.reset."""
    response = self._connection.send(dm_env_rpc_pb2.ResetRequest())
    if self._dm_env_rpc_specs != response.specs:
      raise RuntimeError('Environment changed spec after reset')
    self._last_state = dm_env_rpc_pb2.EnvironmentStateType.INTERRUPTED
    return self.step({})

  def step(self, actions):
    """Implements dm_env.Environment.step."""
    actions = dm_env_flatten_utils.flatten_dict(
        actions, DEFAULT_KEY_SEPARATOR) if self._nested_tensors else actions
    step_response = self._connection.send(
        dm_env_rpc_pb2.StepRequest(
            requested_observations=self._requested_observation_uids,
            actions=self._action_specs.pack(actions)))

    observations = self._observation_specs.unpack(step_response.observations)

    if (step_response.state == dm_env_rpc_pb2.EnvironmentStateType.RUNNING and
        self._last_state == dm_env_rpc_pb2.EnvironmentStateType.RUNNING):
      step_type = dm_env.StepType.MID
    elif step_response.state == dm_env_rpc_pb2.EnvironmentStateType.RUNNING:
      step_type = dm_env.StepType.FIRST
    elif self._last_state == dm_env_rpc_pb2.EnvironmentStateType.RUNNING:
      step_type = dm_env.StepType.LAST
    else:
      raise RuntimeError('Environment transitioned from {} to {}'.format(
          self._last_state, step_response.state))

    self._last_state = step_response.state

    reward = self.reward(
        state=step_response.state,
        step_type=step_type,
        observations=observations)
    discount = self.discount(
        state=step_response.state,
        step_type=step_type,
        observations=observations)
    if not self._is_reward_requested:
      observations.pop(DEFAULT_REWARD_KEY, None)
    if not self._is_discount_requested:
      observations.pop(DEFAULT_DISCOUNT_KEY, None)
    observations = dm_env_flatten_utils.unflatten_dict(
        observations,
        DEFAULT_KEY_SEPARATOR) if self._nested_tensors else observations
    return dm_env.TimeStep(step_type, reward, discount, observations)

  def reward(self, state, step_type, observations):
    """Returns the reward for the given observation state.

    Override in inherited classes to give different reward functions.

    Args:
      state: A dm_env_rpc EnvironmentStateType enum describing the state of the
        environment.
      step_type: The dm_env StepType describing the state of the environment.
      observations: The unpacked observations dictionary mapping string keys to
        scalars and NumPy arrays.

    Returns:
      A reward for the given step.  The shape and type matches that returned by
      `self.reward_spec()`.
    """
    if step_type == dm_env.StepType.FIRST:
      return None
    elif self._default_reward_spec:
      return observations[DEFAULT_REWARD_KEY]
    else:
      return 0.0

  def discount(self, state, step_type, observations):
    """Returns the discount for the given observation state.

    Override in inherited classes to give different discount functions.

    Args:
      state: A dm_env_rpc EnvironmentStateType enum describing the state of the
        environment.
      step_type: The dm_env StepType describing the state of the environment.
      observations: The unpacked observations dictionary mapping string keys to
        scalars and NumPy arrays.

    Returns:
      The discount for the given step.  The shape and type matches that returned
      by `self.discount_spec()`.
    """
    if self._default_discount_spec:
      return observations[DEFAULT_DISCOUNT_KEY]
    if step_type == dm_env.StepType.FIRST:
      return None
    elif (state == dm_env_rpc_pb2.EnvironmentStateType.RUNNING or
          state == dm_env_rpc_pb2.EnvironmentStateType.INTERRUPTED):
      return 1.0
    else:
      return 0.0

  def observation_spec(self):
    """Implements dm_env.Environment.observation_spec."""
    specs = {}
    for uid in self._requested_observation_uids:
      name = self._observation_specs.uid_to_name(uid)
      specs[name] = dm_env_utils.tensor_spec_to_dm_env_spec(
          self._observation_specs.uid_to_spec(uid))
    if not self._is_reward_requested:
      specs.pop(DEFAULT_REWARD_KEY, None)
    if not self._is_discount_requested:
      specs.pop(DEFAULT_DISCOUNT_KEY, None)

    if self._nested_tensors:
      return dm_env_flatten_utils.unflatten_dict(specs, DEFAULT_KEY_SEPARATOR)
    else:
      return specs

  def action_spec(self):
    """Implements dm_env.Environment.action_spec."""
    action_spec = dm_env_utils.dm_env_spec(self._action_specs)
    if self._nested_tensors:
      return dm_env_flatten_utils.unflatten_dict(action_spec,
                                                 DEFAULT_KEY_SEPARATOR)
    else:
      return action_spec

  def reward_spec(self):
    """Implements dm_env.Environment.reward_spec."""
    return (self._default_reward_spec or
            super(DmEnvAdaptor, self).reward_spec())

  def discount_spec(self):
    """Implements dm_env.Environment.discount_spec."""
    return (self._default_discount_spec or
            super(DmEnvAdaptor, self).discount_spec())

  def close(self):
    """Implements dm_env.Environment.close."""
    # Leaves the world if we were joined.  If not, this will be a no-op anyway.
    self._connection.send(dm_env_rpc_pb2.LeaveWorldRequest())
    self._connection = None


def create_world(connection: dm_env_rpc_connection.Connection,
                 create_world_settings: Mapping[str, Any]) -> str:
  """Helper function to create a world with the provided settings.

  Args:
    connection: An instance of Connection already connected to a dm_env_rpc
      server.
    create_world_settings: Settings used to create the world. Values must be
      packable into a Tensor proto or already packed.

  Returns:
    Created world name.
  """

  create_world_settings = {
      key: (value if isinstance(value, dm_env_rpc_pb2.Tensor) else
            tensor_utils.pack_tensor(value))
      for key, value in create_world_settings.items()
  }
  return connection.send(
      dm_env_rpc_pb2.CreateWorldRequest(
          settings=create_world_settings)).world_name


def join_world(
    connection: dm_env_rpc_connection.Connection,
    world_name: str,
    join_world_settings: Mapping[str, Any],
    requested_observations: Optional[Iterable[str]] = None,
    extensions: Optional[Mapping[str, Any]] = immutabledict.immutabledict()
) -> DmEnvAdaptor:
  """Helper function to join a world with the provided settings.

  Args:
    connection: An instance of Connection already connected to a dm_env_rpc
      server.
    world_name: Name of the world to join.
    join_world_settings: Settings used to join the world. Values must be
      packable into a Tensor message or already packed.
    requested_observations: Optional set of requested observations.
    extensions: Optional mapping of extension instances to DmEnvAdaptor
      attributes.

  Returns:
    Instance of DmEnvAdaptor.
  """

  join_world_settings = {
      key: (value if isinstance(value, dm_env_rpc_pb2.Tensor) else
            tensor_utils.pack_tensor(value))
      for key, value in join_world_settings.items()
  }
  specs = connection.send(
      dm_env_rpc_pb2.JoinWorldRequest(
          world_name=world_name, settings=join_world_settings)).specs

  try:
    return DmEnvAdaptor(
        connection, specs, requested_observations, extensions=extensions)
  except ValueError:
    connection.send(dm_env_rpc_pb2.LeaveWorldRequest())
    raise


def create_and_join_world(
    connection: dm_env_rpc_connection.Connection,
    create_world_settings: Mapping[str, Any],
    join_world_settings: Mapping[str, Any],
    requested_observations: Optional[Iterable[str]] = None,
    extensions: Optional[Mapping[str, Any]] = immutabledict.immutabledict()
) -> Tuple[DmEnvAdaptor, str]:
  """Helper function to create and join a world with the provided settings.

  Args:
    connection: An instance of Connection already connected to a dm_env_rpc
      server.
    create_world_settings: Settings used to create the world. Values must be
      packable into a Tensor proto or already packed.
    join_world_settings: Settings used to join the world. Values must be
      packable into a Tensor message.
    requested_observations: Optional set of requested observations.
    extensions: Optional mapping of extension instances to DmEnvAdaptor
      attributes.

  Returns:
    Tuple of DmEnvAdaptor and the created world name.
  """
  world_name = create_world(connection, create_world_settings)
  try:
    return_type = collections.namedtuple('DmEnvAndWorldName',
                                         ['env', 'world_name'])
    return return_type(
        join_world(connection, world_name, join_world_settings,
                   requested_observations, extensions), world_name)
  except (error.DmEnvRpcError, ValueError):
    connection.send(dm_env_rpc_pb2.DestroyWorldRequest(world_name=world_name))
    raise



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""A base class for JoinWorld and LeaveWord tests for a server."""
import abc

from absl.testing import absltest
import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import error
from dm_env_rpc.v1 import tensor_spec_utils


def _find_duplicates(iterable):
  """Returns a list of duplicate entries found in `iterable`."""
  duplicates = []
  seen = set()
  for item in iterable:
    if item in seen:
      duplicates.append(item)
    else:
      seen.add(item)
  return duplicates


def _check_tensor_spec(tensor_spec):
  """Raises an error if the given `tensor_spec` is internally inconsistent."""
  if np.sum(np.asarray(tensor_spec.shape) < 0) > 1:
    raise ValueError(
        f'"{tensor_spec.name}" has shape {tensor_spec.shape} which has more '
        'than one negative element.')
  min_type = tensor_spec.min and tensor_spec.min.WhichOneof('payload')
  max_type = tensor_spec.max and tensor_spec.max.WhichOneof('payload')
  if min_type or max_type:
    _ = tensor_spec_utils.bounds(tensor_spec)


class JoinLeaveWorld(absltest.TestCase, metaclass=abc.ABCMeta):
  """A base class for `JoinWorld` and `LeaveWorld` compliance tests."""

  @property
  def required_join_settings(self):
    """A dict of required settings for a Join World call."""
    return {}

  @property
  def invalid_join_settings(self):
    """A list of dicts of Join World settings which are invalid in some way."""
    return {}

  @abc.abstractproperty
  def world_name(self):
    """A string of the world name of an already created world."""
    pass

  @property
  def invalid_world_name(self):
    """A string which doesn't correspond to any valid world_name."""
    return 'invalid_world_name'

  @property
  @abc.abstractmethod
  def connection(self):
    """An instance of dm_env_rpc's Connection."""
    pass

  def tearDown(self):
    super().tearDown()
    try:
      self.leave_world()
    finally:
      pass

  def join_world(self, **kwargs):
    """Joins the world and returns the spec."""
    response = self.connection.send(dm_env_rpc_pb2.JoinWorldRequest(**kwargs))
    return response.specs

  def leave_world(self):
    """Leaves currently joined world, if any."""
    self.connection.send(dm_env_rpc_pb2.LeaveWorldRequest())

  # pylint: disable=missing-docstring
  def test_can_join(self):
    self.join_world(
        world_name=self.world_name, settings=self.required_join_settings)
    # Success if there's no error raised.

  def test_cannot_join_with_wrong_world_name(self):
    with self.assertRaises(error.DmEnvRpcError):
      self.join_world(world_name=self.invalid_world_name)

  def test_cannot_join_world_with_invalid_settings(self):
    settings = self.required_join_settings
    for name, tensor in self.invalid_join_settings.items():
      with self.assertRaises(error.DmEnvRpcError):
        self.join_world(
            world_name=self.world_name, settings={
                name: tensor,
                **settings
            })

  def test_cannot_join_world_twice(self):
    self.join_world(
        world_name=self.world_name, settings=self.required_join_settings)
    with self.assertRaises(error.DmEnvRpcError):
      self.join_world(
          world_name=self.world_name, settings=self.required_join_settings)

  def test_action_specs_have_unique_names(self):
    specs = self.join_world(
        world_name=self.world_name, settings=self.required_join_settings)
    self.assertEmpty(_find_duplicates(
        spec.name for spec in specs.actions.values()))

  def test_action_specs_for_consistency(self):
    specs = self.join_world(
        world_name=self.world_name, settings=self.required_join_settings)
    for action_spec in specs.actions.values():
      _check_tensor_spec(action_spec)

  def test_observation_specs_have_unique_names(self):
    specs = self.join_world(
        world_name=self.world_name, settings=self.required_join_settings)
    self.assertEmpty(_find_duplicates(
        spec.name for spec in specs.observations.values()))

  def test_observation_specs_for_consistency(self):
    specs = self.join_world(
        world_name=self.world_name, settings=self.required_join_settings)
    for observation_spec in specs.observations.values():
      _check_tensor_spec(observation_spec)

  def test_can_leave_world_if_not_joined(self):
    self.leave_world()
    # Success if there's no error raised.

  def test_can_leave_world_after_joining(self):
    self.join_world(
        world_name=self.world_name, settings=self.required_join_settings)
    self.leave_world()
    # Success if there's no error raised.

  def test_can_rejoin_world_after_leaving(self):
    self.join_world(
        world_name=self.world_name, settings=self.required_join_settings)
    self.leave_world()
    self.join_world(
        world_name=self.world_name, settings=self.required_join_settings)
    # Success if there's no error raised.

  # pylint: enable=missing-docstring



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""A base class for Reset tests for a server."""

import abc

from absl.testing import absltest

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import error


class Reset(absltest.TestCase, metaclass=abc.ABCMeta):
  """A base class for dm_env_rpc `Reset` compliance tests."""

  @property
  @abc.abstractmethod
  def connection(self):
    """An instance of dm_env_rpc's Connection already joined to a world."""
    pass

  @property
  def required_reset_settings(self):
    return {}

  @abc.abstractmethod
  def join_world(self):
    """Joins a world, returning the specs."""
    pass

  def reset(self):
    """Resets the environment, returning the specs."""
    return self.connection.send(dm_env_rpc_pb2.ResetRequest(
        settings=self.required_reset_settings)).specs

  # pylint: disable=missing-docstring
  def test_reset_resends_the_specs(self):
    join_specs = self.join_world()
    specs = self.reset()
    self.assertEqual(join_specs, specs)

  def test_cannot_reset_if_not_joined_to_world(self):
    with self.assertRaises(error.DmEnvRpcError):
      self.reset()

  def test_can_reset_multiple_times(self):
    join_specs = self.join_world()
    self.reset()
    specs = self.reset()
    self.assertEqual(join_specs, specs)
  # pylint: enable=missing-docstring



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""A helper class for sending and receiving property requests and responses.

This helper class provides a Pythonic interface for reading, writing and listing
properties. It simplifies the packing and unpacking of property requests and
responses using the provided dm_env_rpc.v1.connection.Connection instance to
send and receive extension messages.

Example Usage:
  property_extension = PropertyExtension(connection)

  # To read a property:
  value = property_extension['my_property']

  # To write a property:
  property_extension['my_property'] = new_value

  # To find available properties:
  property_specs = property_extension.specs()

  spec = property_specs['my_property']
"""
from typing import Mapping, Sequence, Optional

from dm_env import specs as dm_env_specs
from google.protobuf import any_pb2
from dm_env_rpc.v1 import connection as dm_env_rpc_connection
from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import dm_env_utils
from dm_env_rpc.v1 import tensor_utils
from dm_env_rpc.v1.extensions import properties_pb2


class PropertySpec(object):
  """Class that represents a property's specification."""

  def __init__(self, property_spec_proto: properties_pb2.PropertySpec):
    """Constructs a property specification from PropertySpec proto message.

    Args:
      property_spec_proto: A properties_pb2.PropertySpec message.
    """
    self._property_spec_proto = property_spec_proto

  @property
  def key(self) -> str:
    """Return the property's key."""
    return self._property_spec_proto.spec.name

  @property
  def readable(self) -> bool:
    """Returns True if the property is readable."""
    return self._property_spec_proto.is_readable

  @property
  def writable(self) -> bool:
    """Returns True if the property is writable."""
    return self._property_spec_proto.is_writable

  @property
  def listable(self) -> bool:
    """Returns True if the property is listable."""
    return self._property_spec_proto.is_listable

  @property
  def spec(self) -> Optional[dm_env_specs.Array]:
    """Returns a dm_env spec if the property has a valid dtype.

    Returns:
      Either a dm_env spec or, if the dtype is invalid, None.
    """
    if self._property_spec_proto.spec.dtype != (
        dm_env_rpc_pb2.DataType.INVALID_DATA_TYPE):
      return dm_env_utils.tensor_spec_to_dm_env_spec(
          self._property_spec_proto.spec)
    else:
      return None

  @property
  def description(self) -> str:
    """Returns the property's description."""
    return self._property_spec_proto.description

  def __repr__(self):
    return (f'PropertySpec(key={self.key}, readable={self.readable}, '
            f'writable={self.writable}, listable={self.listable}, '
            f'spec={self.spec}, description={self.description})')


class PropertiesExtension(object):
  """Helper class for sending and receiving property requests and responses."""

  def __init__(self, connection: dm_env_rpc_connection.Connection):
    """Construct extension with provided dm_env_rpc connection to the env.

    Args:
      connection: An instance of Connection already connected to a dm_env_rpc
        server.
    """
    self._connection = connection

  def __getitem__(self, key: str):
    """Alias for PropertiesExtension read function."""
    return self.read(key)

  def __setitem__(self, key: str, value) -> None:
    """Alias for PropertiesExtension write function."""
    self.write(key, value)

  def specs(self, key: str = '') -> Mapping[str, PropertySpec]:
    """Helper to return sub-properties as a dict."""
    return {
        sub_property.key: sub_property for sub_property in self.list(key)
    }

  def read(self, key: str):
    """Reads the value of a property.

    Args:
      key: A string key that represents the property to read.

    Returns:
      The value of the property, either as a scalar (float, int, string, etc.)
      or, if the response tensor has a non-empty `shape` attribute, a NumPy
      array of the payload with the correct type and shape. See
      tensor_utils.unpack for more details.
    """
    response = properties_pb2.PropertyResponse()
    packed_request = any_pb2.Any()
    packed_request.Pack(
        properties_pb2.PropertyRequest(
            read_property=properties_pb2.ReadPropertyRequest(key=key)))
    self._connection.send(packed_request).Unpack(response)
    return tensor_utils.unpack_tensor(response.read_property.value)

  def write(self, key: str, value) -> None:
    """Writes the provided value to a property.

    Args:
      key: A string key that represents the property to write.
      value: A scalar (float, int, string, etc.), NumPy array, or nested lists.
        See tensor_utils.pack for more details.
    """
    packed_request = any_pb2.Any()
    packed_request.Pack(
        properties_pb2.PropertyRequest(
            write_property=properties_pb2.WritePropertyRequest(
                key=key, value=tensor_utils.pack_tensor(value))))
    self._connection.send(packed_request)

  def list(self, key: str = '') -> Sequence[PropertySpec]:
    """Lists properties residing under the provided key.

    Args:
      key: A string key to list properties at this location. If empty, returns
        properties registered at the root level.

    Returns:
      A sequence of PropertySpecs.
    """
    response = properties_pb2.PropertyResponse()
    packed_request = any_pb2.Any()
    packed_request.Pack(
        properties_pb2.PropertyRequest(
            list_property=properties_pb2.ListPropertyRequest(key=key)))
    self._connection.send(packed_request).Unpack(response)

    return tuple(
        PropertySpec(sub_property)
        for sub_property in response.list_property.values)



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""A base class for ResetWorld tests for a server."""

import abc

from absl.testing import absltest

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import error


class ResetWorld(absltest.TestCase, metaclass=abc.ABCMeta):
  """A base class for dm_env_rpc `ResetWorld` compliance tests."""

  @property
  @abc.abstractmethod
  def connection(self):
    """An instance of dm_env_rpc's Connection already joined to a world."""
    pass

  @property
  def required_reset_world_settings(self):
    """Settings necessary to pass to ResetWorld."""
    return {}

  @property
  def required_join_world_settings(self):
    """Settings necessary to pass to JoinWorld."""
    return {}

  @property
  def invalid_world_name(self):
    """The name of a world which doesn't exist."""
    return 'invalid_world_name'

  @property
  @abc.abstractmethod
  def world_name(self):
    """The name of the world to attempt to call ResetWorld on."""
    return ''

  def join_world(self):
    """Joins the world to call ResetWorld on."""
    self.connection.send(dm_env_rpc_pb2.JoinWorldRequest(
        world_name=self.world_name, settings=self.required_join_world_settings))

  def reset_world(self, world_name):
    """Resets the world."""
    self.connection.send(dm_env_rpc_pb2.ResetWorldRequest(
        world_name=world_name, settings=self.required_reset_world_settings))

  def leave_world(self):
    """Leaves the world."""
    self.connection.send(dm_env_rpc_pb2.LeaveWorldRequest())

  # pylint: disable=missing-docstring
  def test_cannot_reset_invalid_world(self):
    with self.assertRaises(error.DmEnvRpcError):
      self.reset_world(self.invalid_world_name)

  def test_can_reset_world_not_joined_to(self):
    self.reset_world(self.world_name)
    # If there are no errors the test passes.

  def test_can_reset_world_when_joined_to_it(self):
    try:
      self.join_world()
      self.reset_world(self.world_name)
      # If there are no errors the test passes.
    finally:
      self.leave_world()
  # pylint: enable=missing-docstring



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for dm_env_rpc/dm_env utilities."""

from absl.testing import absltest
from dm_env import specs
import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import dm_env_utils
from dm_env_rpc.v1 import spec_manager


class TensorSpecToDmEnvSpecTests(absltest.TestCase):

  def test_no_bounds_gives_arrayspec(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.shape[:] = [3]
    tensor_spec.name = 'foo'
    actual = dm_env_utils.tensor_spec_to_dm_env_spec(tensor_spec)
    self.assertEqual(specs.Array(shape=[3], dtype=np.uint32), actual)
    self.assertEqual('foo', actual.name)

  def test_string_give_string_array(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.STRING
    tensor_spec.shape[:] = [1, 2, 3]
    tensor_spec.name = 'string_spec'
    actual = dm_env_utils.tensor_spec_to_dm_env_spec(tensor_spec)
    self.assertEqual(specs.StringArray(shape=[1, 2, 3]), actual)
    self.assertEqual('string_spec', actual.name)

  def test_scalar_with_0_n_bounds_gives_discrete_array(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.name = 'foo'

    max_value = 9
    tensor_spec.min.uint32s.array[:] = [0]
    tensor_spec.max.uint32s.array[:] = [max_value]
    actual = dm_env_utils.tensor_spec_to_dm_env_spec(tensor_spec)
    expected = specs.DiscreteArray(
        num_values=max_value + 1, dtype=np.uint32, name='foo')
    self.assertEqual(expected, actual)
    self.assertEqual(0, actual.minimum)
    self.assertEqual(max_value, actual.maximum)
    self.assertEqual('foo', actual.name)

  def test_scalar_with_1_n_bounds_gives_bounded_array(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.name = 'foo'
    tensor_spec.min.uint32s.array[:] = [1]
    tensor_spec.max.uint32s.array[:] = [10]
    actual = dm_env_utils.tensor_spec_to_dm_env_spec(tensor_spec)
    expected = specs.BoundedArray(
        shape=(), dtype=np.uint32, minimum=1, maximum=10, name='foo')
    self.assertEqual(expected, actual)
    self.assertEqual('foo', actual.name)

  def test_scalar_with_0_min_and_no_max_bounds_gives_bounded_array(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.name = 'foo'
    tensor_spec.min.uint32s.array[:] = [0]
    actual = dm_env_utils.tensor_spec_to_dm_env_spec(tensor_spec)
    expected = specs.BoundedArray(
        shape=(), dtype=np.uint32, minimum=0, maximum=2**32 - 1, name='foo')
    self.assertEqual(expected, actual)
    self.assertEqual('foo', actual.name)

  def test_only_min_bounds(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.shape[:] = [3]
    tensor_spec.name = 'foo'
    tensor_spec.min.uint32s.array[:] = [1]
    actual = dm_env_utils.tensor_spec_to_dm_env_spec(tensor_spec)
    expected = specs.BoundedArray(
        shape=[3], dtype=np.uint32, minimum=1, maximum=2**32 - 1)
    self.assertEqual(expected, actual)
    self.assertEqual('foo', actual.name)

  def test_only_max_bounds(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.shape[:] = [3]
    tensor_spec.name = 'foo'
    tensor_spec.max.uint32s.array[:] = [10]
    actual = dm_env_utils.tensor_spec_to_dm_env_spec(tensor_spec)
    expected = specs.BoundedArray(
        shape=[3], dtype=np.uint32, minimum=0, maximum=10)
    self.assertEqual(expected, actual)
    self.assertEqual('foo', actual.name)

  def test_both_bounds(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.shape[:] = [3]
    tensor_spec.name = 'foo'
    tensor_spec.min.uint32s.array[:] = [1]
    tensor_spec.max.uint32s.array[:] = [10]
    actual = dm_env_utils.tensor_spec_to_dm_env_spec(tensor_spec)
    expected = specs.BoundedArray(
        shape=[3], dtype=np.uint32, minimum=1, maximum=10)
    self.assertEqual(expected, actual)
    self.assertEqual('foo', actual.name)

  def test_bounds_oneof_not_set_gives_dtype_bounds(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.shape[:] = [3]
    tensor_spec.name = 'foo'

    # Just to force the message to get created.
    tensor_spec.min.floats.array[:] = [3.0]
    tensor_spec.min.ClearField('floats')

    actual = dm_env_utils.tensor_spec_to_dm_env_spec(tensor_spec)
    expected = specs.BoundedArray(
        shape=[3], dtype=np.uint32, minimum=0, maximum=2**32 - 1)
    self.assertEqual(expected, actual)
    self.assertEqual('foo', actual.name)

  def test_bounds_wrong_type_gives_error(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.shape[:] = [3]
    tensor_spec.name = 'foo'
    tensor_spec.min.floats.array[:] = [1.9]
    with self.assertRaisesRegex(ValueError, 'uint32'):
      dm_env_utils.tensor_spec_to_dm_env_spec(tensor_spec)

  def test_bounds_on_string_gives_error(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.STRING
    tensor_spec.shape[:] = [2]
    tensor_spec.name = 'named'
    tensor_spec.min.floats.array[:] = [1.9]
    tensor_spec.max.floats.array[:] = [10.0]
    with self.assertRaisesRegex(ValueError, 'string'):
      dm_env_utils.tensor_spec_to_dm_env_spec(tensor_spec)


class DmEnvSpecTests(absltest.TestCase):

  def test_spec(self):
    dm_env_rpc_specs = {
        54:
            dm_env_rpc_pb2.TensorSpec(
                name='fuzz', shape=[3], dtype=dm_env_rpc_pb2.DataType.FLOAT),
        55:
            dm_env_rpc_pb2.TensorSpec(
                name='foo', shape=[2], dtype=dm_env_rpc_pb2.DataType.INT32),
    }
    manager = spec_manager.SpecManager(dm_env_rpc_specs)

    expected = {
        'foo': specs.Array(shape=[2], dtype=np.int32),
        'fuzz': specs.Array(shape=[3], dtype=np.float32)
    }

    self.assertDictEqual(expected, dm_env_utils.dm_env_spec(manager))

  def test_empty_spec(self):
    self.assertDictEqual({},
                         dm_env_utils.dm_env_spec(spec_manager.SpecManager({})))

  def test_spec_generate_and_validate_scalars(self):
    dm_env_rpc_specs = []
    for name, dtype in dm_env_rpc_pb2.DataType.items():
      if dtype != dm_env_rpc_pb2.DataType.INVALID_DATA_TYPE:
        dm_env_rpc_specs.append(
            dm_env_rpc_pb2.TensorSpec(name=name, shape=(), dtype=dtype))

    for dm_env_rpc_spec in dm_env_rpc_specs:
      spec = dm_env_utils.tensor_spec_to_dm_env_spec(dm_env_rpc_spec)
      value = spec.generate_value()
      spec.validate(value)

  def test_spec_generate_and_validate_tensors(self):
    example_shape = (10, 10, 3)

    dm_env_rpc_specs = []
    for name, dtype in dm_env_rpc_pb2.DataType.items():
      if dtype != dm_env_rpc_pb2.DataType.INVALID_DATA_TYPE:
        dm_env_rpc_specs.append(
            dm_env_rpc_pb2.TensorSpec(
                name=name, shape=example_shape, dtype=dtype))

    for dm_env_rpc_spec in dm_env_rpc_specs:
      spec = dm_env_utils.tensor_spec_to_dm_env_spec(dm_env_rpc_spec)
      value = spec.generate_value()
      spec.validate(value)

if __name__ == '__main__':
  absltest.main()
